import os
import copy
from dataclasses import replace
import numpy as np
from dopyqo.units import Unit
import dopyqo

if __name__ == "__main__":
    # A basic config
    config = dopyqo.DopyqoConfig(
        base_folder=os.path.join("qe_files", "Be"),
        prefix="Be",
        active_electrons=2,
        active_orbitals=2,
        run_fci=True,
        n_threads=10,
    )

    # A new config where we want to change the atom positions
    # Besides that it is equal to the config above
    atom_positions = np.array([0.0, 0.5])
    config_new_pos = dopyqo.DopyqoConfig(
        base_folder=os.path.join("qe_files", "Be"),
        prefix="Be",
        active_electrons=2,
        active_orbitals=2,
        run_fci=True,
        n_threads=10,
        atom_positions=atom_positions,
        unit=Unit.CRYSTAL,
    )

    # # Following will fail with a FrozenInstanceError,
    # # since we cannot assign new values to a existing DopyqoConfig.
    # config_new_pos_copy = copy.copy(config)
    # config_new_pos_copy.atom_positions = atom_positions
    # config_new_pos_copy.unit = Unit.CRYSTAL

    # replace will copy the config and assign new values.
    # When using replace all new values will also be checked for their validity.
    # So this is completely equivalent to config_new_pos.
    config_new_pos_replace = replace(
        config,
        atom_positions=atom_positions,
        unit=Unit.CRYSTAL,
    )



import os
import logging
from functools import partial
import multiprocessing
from warnings import warn
from collections import Counter
import numpy as np
from scipy.special import (
    erf,
    spherical_jn,
    sph_harm,
)
from scipy.integrate import simpson
import dopyqo
from dopyqo.pseudopot import Pseudopot
from dopyqo.colors import *


def sph_harm_real(m, n, theta, phi):
    r"""Compute real spherical harmonics.
    The real spherical harmonics are defined as
    $$
    Y_{nm}(\theta, \phi) =
    \begin{cases}
    \sqrt2 (-1)^m \Im{Y_n^{|m|}}&\text{if } m<0\\
    Y_n^0 &\text{if } m=0\\
    \sqrt2 (-1)^m \Re{Y_n^m}&\text{if } m>0
    \end{cases}
    $$
    where $Y_n^m(\theta, \phi)$ are the complex spherical harmonics; see [scipy.special.sph_harm](https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.sph_harm.html).

    Args:
    m: array_like
        Order of the harmonic (int); must have |m| <= n.
    n: array_like
        Degree of the harmonic (int); must have n >= 0. This is often denoted by l (lower case L) in descriptions of spherical harmonics.
    theta: array_like
        Azimuthal (longitudinal) coordinate; must be in [0, 2*pi].
    phi: array_like
        Polar (colatitudinal) coordinate; must be in [0, pi].
    out: ndarray, optional
        Optional output array for the function values

    Returns:
        y_mn: complex scalar or ndarray
            The harmonic sampled at theta and phi.
    """
    if m < 0:
        return np.sqrt(2) * (-1) ** m * sph_harm(m, n, theta, phi).imag
    elif m > 0:
        return np.sqrt(2) * (-1) ** m * sph_harm(m, n, theta, phi).real
    return sph_harm(m, n, theta, phi)


def cart_to_sph(xyz: np.ndarray) -> np.ndarray:
    r"""Transforms cartesian x,y,z coordinates to spherical r, \theta (polar angle), \phi (azimuthal angle) coordinates

    Args:
        xyz (np.ndarray): Array of cartesian coordinates. Shape (N, 3)

    Returns:
        np.ndarray: Array of spherical coordinates. Shape (N, 3)
    """
    ptsnew = np.zeros(xyz.shape)
    xy = xyz[:, 0] ** 2 + xyz[:, 1] ** 2  # x^2+y^2

    # r = sqrt{x^2+y^2+z^2}
    ptsnew[:, 0] = np.sqrt(xy + xyz[:, 2] ** 2)

    # \theta = arctan(\sqrt{x^2+y^2}/z) taking into account the correct quadrant
    # for elevation angle defined from Z-axis down
    ptsnew[:, 1] = np.arctan2(np.sqrt(xy), xyz[:, 2])

    # \phi = arctan(y/x) taking into account the correct quadrant
    # \phi in range [-\pi, \pi]
    ptsnew[:, 2] = np.arctan2(xyz[:, 1], xyz[:, 0])
    # put \phi into range [0, 2\pi], by adding 2\pi to negative angles
    ptsnew[:, 2] = np.where(ptsnew[:, 2] < 0.0, ptsnew[:, 2] + 2 * np.pi, ptsnew[:, 2])
    return ptsnew


def v_loc_pw(
    p: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    pseudopot: Pseudopot,
) -> np.ndarray:
    r"""Calculates the matrix of the local pseudopotential V_loc(p, p') in the plane wave basis in Hartree atomic units

    V_loc(p, p') = 4\pi/V \int_0^\infty dr r (V_loc(r)+Z erf(r)/r) sin((p-p')r)/(p-p') - 4\pi/V Z e^{-(p-p')^2/4}/(p-p')^2
    which is the Fourier transform of V_loc(r) + Z erf(r)/r - Z erf(r)/r

    A term -Z erf(r)/r is subtracted (-[-Z erf(r)/r]=+Z erf(r)/r) in real space (thus making the function short-ranged)
    and added again in G space
    V_loc^short(p, p') = 4\pi/V \int_0^\infty dr r (V_loc(r)+Z erf(r)/r) sin((p-p')r)/(p-p')
    which is continious for p-p'=0 but the p-p'=0 term is calculated differently (see below)

    The Fourier transform of Z erf(r)/r is
    4\pi/V Z e^{-(p-p')^2/4}/(p-p')^2

    The p-p'=0 limit for V_loc is the p-p'=0 limit of
    4\pi/V \int_0^\infty dr r^2 (V_loc(r)+Z/r)
    since the p-p'=0 limit for the -Z/r part of V_loc is cancelled by the electronic background
    Because V_loc does not behave exactly like -Z/r, we still have to calculate the p-p'=0 limit
    for V_loc+Z/r

    As a reference see the implementation in Quantum ESPRESSO in its soruce q-e/upflib/vloc_mod.f90

    Args:
        p (np.ndarray): reciprocal space vectors. Shape (#waves, 3)
        cell_volume (float): Volume V of the real-space unit cell
        atom_positions (np.ndarray): Array of the coordinates R_I of every atom described
                                     by the pseudopotential. Shape (#atoms, 3)
        pseudopot (Pseudopot): Pseudopotential object containing z_valence, V_loc(r), and the radial grid

    Returns:
        np.ndarray:  V_loc(p, p')
                     where p, p' take all values from input array p
    """
    # /0.5 to convert from Hartree to Rydberg since the PP is given in Rydberg atomic units
    z_valence = pseudopot.z_valence / 0.5
    v_loc_r = pseudopot.pp_local
    r_grid = pseudopot.r_grid

    p_prime = p
    p_minus_p_prime = p[:, None] - p_prime[None]  # shape (#waves, #waves, 3)
    # p_prime[None] is of shape (1, #waves, 3) while p[:, None] is of shape (#waves, 1, 3)
    # so p_minus_p_prime[i,j] = p[i] - p_prime[j]

    p_minus_p_prime_dot_R = np.sum(
        p_minus_p_prime[None] * atom_positions[:, None, None], axis=3
    )  # sum over 3D-coordinates, shape (#atoms, #waves, #waves)
    # \sum_I 4\pi/V e^{-i (p-p').R_I}
    prefactor = 4 * np.pi / cell_volume * np.sum(np.exp(-1j * p_minus_p_prime_dot_R), axis=0)  # sum over atoms, shape (#waves, #waves)

    # Rydberg to Hartree since the PP is given in Rydberg atomic units
    prefactor = prefactor / 2

    # Fourier transform of V_loc(r) + Z erf(r)/r (short-range part of V_loc) is
    # 4\pi/V \int dr r (V_loc(r)+Z erf(r)/r) sin((p-p')r)/(p-p')
    p_norms = np.linalg.norm(p_minus_p_prime, ord=2, axis=-1)
    p_norms_unique, unique_inv = np.unique(p_norms, return_inverse=True)
    # p_norms obviously has a lot of repeating values, only calculate V_loc for unique p_norms
    # Now p_norms.flatten() is equal to p_norms_unique[unique_inv])
    # consequently p_norms is equal to p_norms_unique[unique_inv].reshape(p_norms.shape)

    integrands = np.zeros(p_norms_unique.shape + (len(r_grid.r),), dtype=np.float64)
    mask = p_norms_unique > 1e-8
    # removal of erf(r)/r term
    integrands[mask, :] = (
        (r_grid.r * v_loc_r + z_valence * erf(r_grid.r))
        * np.sin(p_norms_unique[mask][:, np.newaxis] * r_grid.r)
        / p_norms_unique[mask][:, np.newaxis]
    )
    # p_norm = 0 limit, since it is continuous
    # integrands[~mask, :] = r_grid.r * (r_grid.r * v_loc_r + z_valence * erf(r_grid.r))
    f_v_plus_erf = simpson(y=integrands, x=r_grid.r)

    # Fourier transform of Z erf(r)/r is 4\pi/V Z e^{-(p-p')^2/4}/(p-p')^2 which is added in G-space
    f_erf = np.zeros_like(p_norms_unique, dtype=np.complex128)  # set G=0 to zero
    f_erf[mask] = z_valence * np.exp(-p_norms_unique[mask] ** 2 / 4) / p_norms_unique[mask] ** 2

    f_v = f_v_plus_erf - f_erf

    # Fourier transform of V_loc(r)+Z/r for G=0
    p_zero_lim = simpson(y=r_grid.r * (r_grid.r * v_loc_r + z_valence), x=r_grid.r)
    p_zero_lim_mat = np.zeros_like(p_norms_unique, dtype=np.complex128)  # set G!=0 to zero
    p_zero_lim_mat[~mask] = p_zero_lim
    # print(f"p_zero_lim: {p_zero_lim}")

    v_loc_mat = f_v + p_zero_lim_mat
    # for i, val in enumerate(v_loc_mat):
    #     print(
    #         f"i: {i+1}\ngl: {p_norms_unique[i]**2}, vloc: {4*np.pi/cell_volume*val}\n"
    #     )

    return prefactor * (v_loc_mat[unique_inv].reshape(p_norms.shape))


def v_loc(
    p: np.ndarray,
    c_ip: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    pseudopot: Pseudopot,
    save_filename_pw: str | None = None,
) -> np.ndarray:
    r"""Calculates the matrix of the local pseudopotential V_loc(i, i) in the Kohn-Sham basis
    V_loc(i, j) = \sum_{p, p'} c_{ip} V_loc(p, p') c_{jp'}

    See v_loc_pw for the calculation of V_loc(p, p')

    Args:
        p (np.ndarray): reciprocal space vectors. Shape (#waves, 3)
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis
        cell_volume (float): Volume V of the real-space unit cell
        atom_positions (np.ndarray): Array of the coordinates R_I of every atom described
                                     by the pseudopotential. Shape (#atoms, 3)
        pseudopot (Pseudopot): Pseudopotential object containing z_valence, V_loc(r), and the radial grid

    Returns:
        np.ndarray:  V_loc(i, j') = \sum_{p,p'} c_{p,i}^* c_{p',j} V_loc(p, p')
                     where p, p' take all values from input array p
                     where i, i' represent Kohn-Sham band indices
    """
    v_loc_pw_mat = v_loc_pw(p, cell_volume, atom_positions, pseudopot)
    if save_filename_pw is not None:
        np.save(save_filename_pw, v_loc_pw_mat)
    return c_ip.conj() @ v_loc_pw_mat @ c_ip.T


def v_nl_pw(
    p: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    pseudopot: Pseudopot,
) -> np.ndarray:
    r"""Calculates the matrix of the non-local pseudopotential in the plane wave basis in Hartree atomic units
    V_nl(p, p') = (4\pi)^2/V \sum_I e^(-i (p-p') . R_I) \sum_{lm} \sum_{ij} Y_{lm}(p) Y*_{lm}(p') F^i_l(p) D_{ij} F^j_l(p')
                = (4\pi)^2/V \sum_I e^(-i (p-p') . R_I) \sum_{ij} D_{ij} \sum_l [\sum_m Y_{lm}(p) Y*_{lm}(p')] F^i_l(p) F^j_l(p')
    where
    F^i_l(p) = \int dr r^2 \beta^i_l(r) j_l(pr)
    where j_l(x) is the spherical Bessel function

    As a reference see https://docs.abinit.org/theory/pseudopotentials/

    Args:
        p (np.ndarray): reciprocal space vectors. Shape (#waves, 3)
        cell_volume (float): Volume V of the real-space unit cell
        atom_positions (np.ndarray): Array of the coordinates R_I of every atom described
                                     by the pseudopotential. Shape (#atoms, 3)
        pseudopot (Pseudopot): Pseudopotential object containing D_ij, \beta^i_l(r), and the radial grid

    Returns:
        np.ndarray: V_nl(p, p')
                    where p, p' take all values from input array p and
    """
    d_ij = pseudopot.pp_dij
    beta_projectors = pseudopot.pp_betas
    r_grid = pseudopot.r_grid

    assert len(beta_projectors) == len(np.unique([x.idx for x in beta_projectors])), "Beta projectors are invalid! Their indices are not unique."
    beta_projectors = sorted(beta_projectors, key=lambda x: x.idx)

    # Check that D_ij does not connect beta projectors of different angular momentum
    assert (
        d_ij.shape == (len(beta_projectors),) * 2
    ), f"Number of beta projectors ({len(beta_projectors)}) does not match shape of D_ij ({d_ij.shape})."
    for i, bp_i in enumerate(beta_projectors):
        for j, bp_j in enumerate(beta_projectors):
            if bp_i.angular_momentum != bp_j.angular_momentum:
                assert np.isclose(d_ij[i, j], 0.0), (
                    f"D_ij at i={i}, j={j} (D_{i},{j}={d_ij[i,j]}) connects beta projectors "
                    + f"of different angular momentum, l_i={bp_i.angular_momentum} and l_j={bp_j.angular_momentum}, "
                    + "which should not be the case."
                )

    p_prime = p

    p_minus_p_prime = p[:, None] - p_prime[None]  # shape (#waves, #waves, 3)

    p_minus_p_prime_dot_R = np.sum(
        p_minus_p_prime[None] * atom_positions[:, None, None], axis=3
    )  # sum over 3D-coordinates, shape (#atoms, #waves, #waves)
    # (4\pi)^2/V \sum_I e^(-i (p-p') . R_I)
    prefactor = (4 * np.pi) ** 2 / cell_volume * np.sum(np.exp(-1j * p_minus_p_prime_dot_R), axis=0)  # sum over atoms, shape (#waves, #waves)

    # Rydberg to Hartree since the PP is given in Rydberg atomic units
    prefactor = prefactor / 2

    # Set negative y-axis to axis from which to measure polar angle, instead of z-axis
    # x, y, z -> x, z, -y
    # To make coordinate system right-handed again, set x -> -x
    # Use coordinates -x, z, -y
    # This is the convention used in Quantum ESPRESSO and reproduces
    # the non-local pp matrix elements calculated in Quantum ESPRESSO for Be
    # see QE source code in upflib/init_us_2_acc.f90 line 112 and following,
    # especially the variable vkb_:
    #   vkb = (4pi/sqrt(omega)).Y_lm(q).f_l(q).(i^l).S(q)
    #   vq = f_l(q) = 4\pi/sqrt(omega) int r beta(r) j_l(qr)
    #   ylm are real spherical harmonics in the above described coordinate system.
    #       Order of m values in ylm for l=1: -1, +1, 0
    #   pref = (-i)^l
    #   sk = e^(-i * p.R) where R is the position of one atom
    # NOTE: This coordinate transformation still results in a right-handed
    #       coordinate system and does not change the energies of the Hamiltonian
    p_cart = p.copy()
    p_cart = p_cart[:, [0, 2, 1]]  # xyz to xzy
    p_cart[:, 0] *= -1  # to -x, z, y
    p_cart[:, 2] *= -1  # to -x, z, -y

    p_sph = cart_to_sph(p_cart)
    p_r = p_sph[:, 0]  # Radii p vectors
    p_theta = p_sph[:, 1]  # Theta (polar) angles of p vectors
    p_phi = p_sph[:, 2]  # Phi (azimuthal) angles of p vectors

    # f_il_p = f_i(p) = \int dr r^2 \beta^i_l(r) j_l(pr)
    # Note that l is uniquely defined by i, so l=l_i and
    # f_il_p only depends on i and p
    f_il_p = np.zeros((len(beta_projectors), len(p)))
    for i, bp in enumerate(beta_projectors):
        # \int dr r^2 \beta^i_l(r) j_l(Gr), where bp.values is r \beta^i_l(r)
        integrands = [r_grid.r * bp.values * spherical_jn(n=bp.angular_momentum, z=p_r_val * r_grid.r) for p_r_val in p_r]

        f_il_p[i] = simpson(y=integrands, x=r_grid.r)

    # Precompute spherical harmonics Y_{lm}(p) and Y*_{lm}(p')
    y_lm_p = {}
    for l_val in np.unique([bp.angular_momentum for bp in beta_projectors]):
        # There are different conventions for the meanings of the input arguments theta and phi.
        # In SciPy theta is the azimuthal angle and phi is the polar angle.
        # It is common to see the opposite convention, that is, theta as the polar angle
        # and phi as the azimuthal angle.
        # logging.info(f"m takes values {list(range(-l, l + 1))} for l={l}")

        # Complex spherical harmonics
        # y_lm_g[l_val] = np.array(
        #     [sph_harm(m, l_val, p_phi, p_theta) for m in range(-l_val, l_val + 1)]
        # )
        # !!!: sph_harm_real is deprecated since scipy version 1.15.0:
        #      "This function is deprecated and will be removed in SciPy 1.17.0.
        #           Please use scipy.special.sph_harm_y instead."
        #      This code was written for scipy version 1.13.0.
        # !!!: scipy.special.sph_harm_y redefines theta and phi:
        #      theta is the polar angle and phi is the azimuthal angle
        #      For sph_harm_real:
        #      theta is the azimuthal angle and phi is the polar angle

        # NOTE: Real spherical harmonics can be used for Hamiltonians without spin-orbit coupling
        # since then the Hamiltonian and the energies do not depend on the l and m quantum numbers
        y_lm_p[l_val] = np.array([sph_harm_real(m, l_val, p_phi, p_theta) for m in range(-l_val, l_val + 1)])

    v_nl_mat = np.zeros((len(p), len(p_prime)), dtype=np.complex128)

    # m_sums(l)(p, p') = \sum_m Y_lm(p) Y*_lm(p')
    m_sums = {key: np.sum(val[:, :, None] * val[:, None, :].conj(), axis=0) for key, val in y_lm_p.items()}

    for i, bp_i in enumerate(beta_projectors):
        assert bp_i.idx - 1 == i
        l_i = bp_i.angular_momentum
        f_il_p_i = f_il_p[i, :, None]

        for j, bp_j in enumerate(beta_projectors):
            assert bp_j.idx - 1 == j
            l_j = bp_j.angular_momentum
            f_jl_p_j = f_il_p[j, None, :]

            if l_i != l_j:
                continue

            v_nl_mat += d_ij[i, j] * (f_il_p_i @ f_jl_p_j) * m_sums[l_i]

    return prefactor * v_nl_mat


def v_nl(
    p: np.ndarray,
    c_ip: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    pseudopot: Pseudopot,
    save_filename_pw: str | None = None,
) -> np.ndarray:
    r"""Calculates the matrix of the non-local pseudopotential in the Kohn-Sham basis in Hartree atomic units
    V_nl(i,j) = \sum_{p, p'} c_{ip} V_nl(p, p') c_{jp'}

    See v_nl_pw for the calculation of V_nl(p, p')

    Args:
        p (np.ndarray): reciprocal space vectors. Shape (#waves, 3)
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis
        cell_volume (float): Volume V of the real-space unit cell
        atom_positions (np.ndarray): Array of the coordinates R_I of every atom described
                                     by the pseudopotential. Shape (#atoms, 3)
        pseudopot (Pseudopot): Pseudopotential object containing D_ij, \beta^i_l(r), and the radial grid

    Returns:
        np.ndarray: V_nl(i, j') = \sum_{p,p'} c_{p,i}^* c_{p',j} V_nl(p, p')
                    where p, p' take all values from input array p and
                    where i, i' represent Kohn-Sham band indices
    """
    v_nl_pw_mat = v_nl_pw(p, cell_volume, atom_positions, pseudopot)
    if save_filename_pw is not None:
        np.save(save_filename_pw, v_nl_pw_mat)
    return c_ip.conj() @ v_nl_pw_mat @ c_ip.T


def calc_pps(
    p: np.ndarray,
    c_ip: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    pseudopots: list[Pseudopot],
    save_filename: str | None = None,
    rust_impl: bool = True,
    n_threads: int = 1,
) -> np.ndarray:
    r"""Calculate sum of all pseudopotentials in Hartree atomic units.
    Calls v_nl and v_loc after checking arguments.

    Args:
        p (np.ndarray): Array of momentum vectors. Shape (#waves, 3)
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis. Shape (#bands, #waves)
        atom_positions (np.ndarray): 2D array of positions of each atom. Shape (#atoms, 3)
        atomic_numbers (np.ndarray): 1D array of atomic number of each atom. Shape (#atoms,)
        cell_volume (float): Cell volume of the computationen real space cell.
        pseudopot (list[Pseudopot]): Pseudopotential objects for all atom types

    Returns:
        np.ndarray: Nuclear interaction matrix
    """
    n_threads = int(n_threads)
    assert n_threads > 0, f"Number of threads needs to be positive but is {n_threads}!"
    if rust_impl:
        try:
            import dopyqo_rs as calc_rs
        except ImportError:
            print(f"{ORANGE}Import warning: Could not import dopyqo_rs package. Falling back to python implementation.{RESET_COLOR}")
            v_loc_pw_func = v_loc_pw
            v_nl_pw_func = v_nl_pw
            rust_impl = False  # Set such that variable accurately represents if rust implementation is used
        else:  # No exception
            v_loc_pw_func = None
            v_nl_pw_func = None
            if n_threads > 1:
                v_loc_func = partial(calc_rs.v_loc_par, n_threads=n_threads)
                v_nl_func = partial(calc_rs.v_nl_par, n_threads=n_threads)
                logging.info(f"Using Rust implementation with {n_threads} threads.")
            else:
                v_loc_func = calc_rs.v_loc
                v_nl_func = calc_rs.v_nl
                logging.info("Using Rust implementation.")
            if save_filename is not None:
                logging.info(f"\tPseudopotential will not be saved to {save_filename}!")
    else:
        v_loc_pw_func = v_loc_pw
        v_nl_pw_func = v_nl_pw

    # --------------------------------- CHECKING FOR INVALID INPUTS ---------------------------------
    assert c_ip.shape[1] == p.shape[0], f"c_ip and p arrays have different number of plane waves ({c_ip.shape[1]} vs. {p.shape[0]})!"

    assert atom_positions.shape[0] == len(atomic_numbers), (
        "Atomic numbers array contains different number of atoms than atom positions array "
        + f"({len(atomic_numbers)} vs. {atom_positions.shape[0]})!"
    )

    # Mapping atomic numbers to list of atom positions
    atoms_dict = {num: [] for num in atomic_numbers}
    for i, atomic_num in enumerate(atomic_numbers):
        atoms_dict[atomic_num].append(atom_positions[i])
    atoms_dict = {key: np.array(val) for key, val in atoms_dict.items()}
    pp_dict = {pp.atomic_number: pp for pp in pseudopots}

    # Checking for duplicate PPs
    atomic_nums_pp_duplicates = [
        atomic_num_pp for atomic_num_pp, count_val in Counter([pp.atomic_number for pp in pseudopots]).items() if count_val > 1
    ]
    atomic_el_pp_duplicates = [dopyqo.elements_to_atomic_number[x] for x in atomic_nums_pp_duplicates]
    assert len(atomic_nums_pp_duplicates) == 0, "More than one pseudopotential given for atoms " + ", ".join(
        str(x) + f" (Z={atomic_nums_pp_duplicates[i]})" for i, x in enumerate(atomic_el_pp_duplicates)
    )

    # Checking for atoms without PP
    atomic_num_wo_pp = [atomic_num for atomic_num in atoms_dict.keys() if atomic_num not in pp_dict.keys()]
    atomic_el_wo_pp = [dopyqo.elements_to_atomic_number[x] for x in atomic_num_wo_pp]
    assert len(atomic_num_wo_pp) == 0, "No pseudopotentials given for atoms " + ", ".join(
        str(x) + f" (Z={atomic_num_wo_pp[i]})" for i, x in enumerate(atomic_el_wo_pp)
    )

    # Delete all PPs with no corresponding atoms
    pp_dict = {key: val for key, val in pp_dict.items() if key in atoms_dict.keys()}

    ############################ CALCULATING PPs ############################
    v_nl_mat = 0.0
    v_loc_mat = 0.0
    for atomic_num, pp in pp_dict.items():
        logging.info(
            "Calculating PP for element %s (Z=%i)...",
            dopyqo.elements_to_atomic_number[atomic_num],
            atomic_num,
        )
        pos = atoms_dict[atomic_num]
        # logging.info(f"Atomic positions: {pos}")
        logging.info("Calculating local PP...")
        if rust_impl:
            v_loc_mat += v_loc_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_loc_mat += v_loc_pw_func(p, cell_volume, pos, pp)
        logging.info("Calculating non-local PP...")
        if rust_impl:
            v_nl_mat += v_nl_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_nl_mat += v_nl_pw_func(p, cell_volume, pos, pp)

    v_pp = v_loc_mat + v_nl_mat

    if save_filename is not None and not rust_impl:
        save_folder = os.path.join(*os.path.split(save_filename)[:-1])
        os.makedirs(save_folder, exist_ok=True)
        np.save(os.path.join(save_filename), v_pp)

    if rust_impl:
        return v_pp

    return c_ip.conj() @ v_pp @ c_ip.T


def calc_pps_concurrent(
    p: np.ndarray,
    c_ip: np.ndarray,
    cell_volume: float,
    atom_positions: np.ndarray,
    atomic_numbers: np.ndarray,
    pseudopots: list[Pseudopot],
    save_filename: str | None = None,
    rust_impl: bool = True,
) -> np.ndarray:
    r"""Calculate sum of all pseudopotentials in Hartree atomic units.
    Calls v_nl and v_loc after checking arguments.

    Args:
        p (np.ndarray): Array of momentum vectors
        c_ip (np.ndarray): Array of coefficients describing the Kohn-Sham orbitals in the plane wave basis
        atom_positions (np.ndarray): 2D array of positions of each atom. Shape (#atoms, 3)
        atomic_numbers (np.ndarray): 1D array of atomic number of each atom. Shape (#atoms,)
        cell_volume (float): Cell volume of the computationen real space cell.
        pseudopot (list[Pseudopot]): Pseudopotential objects for all atom types

    Returns:
        np.ndarray: Nuclear interaction matrix
    """
    if rust_impl:
        try:
            import dopyqo_rs as calc_rs
        except ImportError:
            warn("Could not import dopyqo_rs package. Falling back to python implementation.")
            v_loc_pw_func = v_loc_pw
            v_nl_pw_func = v_nl_pw
            rust_impl = False  # Set such that variable accurately represents if rust implementation is used
        else:  # No exception
            v_loc_pw_func = None
            v_nl_pw_func = None
            v_loc_func = calc_rs.v_loc
            v_nl_func = calc_rs.v_nl
            logging.info("Using Rust implementation.")
            if save_filename is not None:
                logging.info(f"\tPseudopotential will not be saved to {save_filename}!")
    else:
        v_loc_pw_func = v_loc_pw
        v_nl_pw_func = v_nl_pw

    # --------------------------------- CHECKING FOR INVALID INPUTS ---------------------------------
    assert c_ip.shape[1] == p.shape[0], f"c_ip and p arrays have different number of plane waves ({c_ip.shape[1]} vs. {p.shape[0]})!"

    assert atom_positions.shape[0] == len(atomic_numbers), (
        "Atomic numbers array contains different number of atoms than atom positions array "
        + f"({len(atomic_numbers)} vs. {atom_positions.shape[0]})!"
    )

    # Mapping atomic numbers to list of atom positions
    atoms_dict = {num: [] for num in atomic_numbers}
    for i, atomic_num in enumerate(atomic_numbers):
        atoms_dict[atomic_num].append(atom_positions[i])
    atoms_dict = {key: np.array(val) for key, val in atoms_dict.items()}
    pp_dict = {pp.atomic_number: pp for pp in pseudopots}

    # Checking for duplicate PPs
    atomic_nums_pp_duplicates = [
        atomic_num_pp for atomic_num_pp, count_val in Counter([pp.atomic_number for pp in pseudopots]).items() if count_val > 1
    ]
    atomic_el_pp_duplicates = [dopyqo.elements_to_atomic_number[x] for x in atomic_nums_pp_duplicates]
    assert len(atomic_nums_pp_duplicates) == 0, "More than one pseudopotential given for atoms " + ", ".join(
        str(x) + f" (Z={atomic_nums_pp_duplicates[i]})" for i, x in enumerate(atomic_el_pp_duplicates)
    )

    # Checking for atoms without PP
    atomic_num_wo_pp = [atomic_num for atomic_num in atoms_dict.keys() if atomic_num not in pp_dict.keys()]
    atomic_el_wo_pp = [dopyqo.elements_to_atomic_number[x] for x in atomic_num_wo_pp]
    assert len(atomic_num_wo_pp) == 0, "No pseudopotentials given for atoms " + ", ".join(
        str(x) + f" (Z={atomic_num_wo_pp[i]})" for i, x in enumerate(atomic_el_wo_pp)
    )

    # Delete all PPs with no corresponding atoms
    pp_dict = {key: val for key, val in pp_dict.items() if key in atoms_dict.keys()}

    ############################ CALCULATING PPs ############################
    # Helper function to calculate local and non-local potentials (with multiprocessing)
    def calculate_loc(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, queue):
        logging.info(
            "Calculating PP for element %s (Z=%i)...",
            dopyqo.elements_to_atomic_number[atomic_num],
            atomic_num,
        )
        pos = atoms_dict[atomic_num]
        logging.info(f"Atomic positions: {pos}")

        # Calculate local PP
        logging.info("Calculating local PP...")
        if rust_impl:
            v_loc_mat = v_loc_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_loc_mat = v_loc_pw_func(p, cell_volume, pos, pp)

        queue.put(v_loc_mat)

    def calculate_nl(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, queue):
        logging.info(
            "Calculating PP for element %s (Z=%i)...",
            dopyqo.elements_to_atomic_number[atomic_num],
            atomic_num,
        )
        pos = atoms_dict[atomic_num]
        logging.info(f"Atomic positions: {pos}")

        # Calculate non-local PP
        logging.info("Calculating non-local PP...")
        if rust_impl:
            v_nl_mat = v_nl_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_nl_mat = v_nl_pw_func(p, cell_volume, pos, pp)

        queue.put(v_nl_mat)

    def calculate_pp(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, queue):
        logging.info(
            "Calculating PP for element %s (Z=%i)...",
            dopyqo.elements_to_atomic_number[atomic_num],
            atomic_num,
        )
        pos = atoms_dict[atomic_num]
        logging.info(f"Atomic positions: {pos}")

        v_loc_mat = 0  # Initialize local matrix for this process
        v_nl_mat = 0  # Initialize non-local matrix for this process

        # Calculate local PP
        logging.info("Calculating local PP...")
        if rust_impl:
            v_loc_mat += v_loc_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_loc_mat += v_loc_pw_func(p, cell_volume, pos, pp)

        # Calculate non-local PP
        logging.info("Calculating non-local PP...")
        if rust_impl:
            v_nl_mat += v_nl_func(p, c_ip, cell_volume, pos, pp)
        else:
            v_nl_mat += v_nl_pw_func(p, cell_volume, pos, pp)

        # return v_loc_mat, v_nl_mat
        queue.put(v_loc_mat + v_nl_mat)

    # data_queue = [multiprocessing.Queue() for _ in range(2 * len(pp_dict.keys()))]
    data_queue = [multiprocessing.Queue() for _ in range(len(pp_dict.keys()))]
    process_list = []

    for i, (atomic_num, pp) in enumerate(pp_dict.items()):
        # process_loc = multiprocessing.Process(
        #     target=calculate_loc, args=(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, data_queue[2 * i])
        # )
        # process_loc.start()
        # process_list.append(process_loc)
        # process_nl = multiprocessing.Process(
        #     target=calculate_nl, args=(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, data_queue[2 * i + 1])
        # )
        # process_nl.start()
        # process_list.append(process_nl)

        process = multiprocessing.Process(target=calculate_pp, args=(atomic_num, pp, atoms_dict, p, c_ip, cell_volume, rust_impl, data_queue[i]))
        process.start()
        process_list.append(process)

    for proc in process_list:
        proc.join()

    data = [q_data for q in data_queue if (q_data := q.get())[1] is not None]
    v_pp = sum(data)

    if save_filename is not None and not rust_impl:
        save_folder = os.path.join(*os.path.split(save_filename)[:-1])
        os.makedirs(save_folder, exist_ok=True)
        np.save(os.path.join(save_filename), v_pp)

    if rust_impl:
        return v_pp

    return c_ip.conj() @ v_pp @ c_ip.T


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from wfc import Wfc

    logging.basicConfig(level=logging.INFO)

    # We first investigate fully non-local norm-conserving pseudopotentials.
    # This is specified by pseudo_type = "NC" in the upf-file.

    # See equations (3.18) and onwards in "Planewaves, Pseudopotentials,
    # and the LAPW Method: Second Edition" from David J. Singh and Lars Nordstrom
    # for an explanation of the local and non-local part of the Kleinman-Bylander
    # transformed pseudopotential

    # Further see https://docs.abinit.org/theory/pseudopotentials/
    # and https://www.tcm.phy.cam.ac.uk/~jry20/gipaw/tutorial_pp.pdf

    base_folder = os.path.join("qe_files", "H2")
    dat_file = os.path.join(base_folder, "H2.save", "wfc1.dat")
    xml_file = os.path.join(base_folder, "H2.save", "data-file-schema.xml")

    base_folder = os.path.join("qe_files", "He")
    dat_file = os.path.join(base_folder, "He.save", "wfc1.dat")
    xml_file = os.path.join(base_folder, "He.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "H2_15")
    # dat_file = os.path.join(base_folder, "H2.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "H2.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "H2_pos_-1.9")
    # dat_file = os.path.join(base_folder, "H2.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "H2.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "LiH")
    # dat_file = os.path.join(base_folder, "LiH.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "LiH.save", "data-file-schema.xml")

    base_folder = os.path.join("qe_files", "Be")
    dat_file = os.path.join(base_folder, "Be.save", "wfc1.dat")
    xml_file = os.path.join(base_folder, "Be.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "BeH")
    # dat_file = os.path.join(base_folder, "BeH.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "BeH.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "BeH_2")
    # dat_file = os.path.join(base_folder, "BeH.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "BeH.save", "data-file-schema.xml")

    # base_folder = os.path.join("qe_files", "Mg")
    # dat_file = os.path.join(base_folder, "Mg.save", "wfc1.dat")
    # xml_file = os.path.join(base_folder, "Mg.save", "data-file-schema.xml")

    # Active space selection and options
    active_electrons = 2
    active_orbitals = 2
    binary_occupations = False

    print("Reading data from files...")
    wfc1_ncpp = Wfc.from_file(dat_file, xml_file)
    # wfc1_ncpp.atomic_numbers = wfc1_ncpp.atomic_numbers[:1]
    # wfc1_ncpp.atom_positions_hartree = wfc1_ncpp.atom_positions_hartree[:1]

    # wfc1_ncpp.atomic_numbers = wfc1_ncpp.atomic_numbers[1:]
    # wfc1_ncpp.atom_positions_hartree = wfc1_ncpp.atom_positions_hartree[1:]

    # wfc1_ncpp.atomic_numbers = np.array([1])
    # wfc1_ncpp.atom_positions_hartree = np.array([[0.5, 0.5, 0.5]])
    # wfc1_ncpp.cell_volume = 1

    wfc1_ncpp.k_plus_G = wfc1_ncpp.k_plus_G[:10]
    wfc1_ncpp.evc = np.eye(wfc1_ncpp.k_plus_G.shape[0])

    pp_base_folder = os.path.join(os.path.expanduser("~"), "Pseudopotentials")
    xml_file_H = os.path.join(pp_base_folder, "H_ONCV_PBE-1.2.upf")
    xml_file_He = os.path.join(pp_base_folder, "He_ONCV_PBE-1.2.upf")
    xml_file_Li = os.path.join(pp_base_folder, "Li_ONCV_PBE-1.2.upf")
    xml_file_Be = os.path.join(pp_base_folder, "Be_ONCV_PBE-1.2.upf")
    xml_file_Mg = os.path.join(pp_base_folder, "Mg_ONCV_PBE-1.2.upf")

    pp_H = Pseudopot(xml_file_H)
    pp_He = Pseudopot(xml_file_He)
    pp_Li = Pseudopot(xml_file_Li)
    pp_Be = Pseudopot(xml_file_Be)
    pp_Mg = Pseudopot(xml_file_Mg)

    pseudopots = [pp_H, pp_He, pp_Li, pp_Be, pp_Mg]

    pp = pp_He

    v_nl_r = np.zeros_like(pp.pp_mesh_r)
    for i in range(pp.pp_nonlocal["PP_DIJ"].shape[0]):
        for j in range(pp.pp_nonlocal["PP_DIJ"].shape[1]):
            v_nl_r += (
                pp.pp_nonlocal["PP_DIJ"][i, j]
                * pp.pp_nonlocal[f"PP_BETA.{i+1}"]
                / np.array(pp.pp_mesh_r)
                * pp.pp_nonlocal[f"PP_BETA.{j+1}"]
                / np.array(pp.pp_mesh_r)
            )

    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(7, 6), constrained_layout=True)
    # local part, long-ranged and behaving like -Z/r for r \to \infty
    ax[0][0].plot(pp.pp_mesh_r, pp.pp_local, label="local")
    cut = 40
    # -Z/r; /0.5 to convert from Hartree to Rydberg
    ax[0][0].plot(pp.pp_mesh_r[cut:], -pp.z_valence / pp.pp_mesh_r[cut:] / 0.5, "--", label="-Z/r")
    ax[0][0].plot(
        pp.pp_mesh_r,
        pp.pp_local + pp.z_valence * erf(pp.pp_mesh_r) / pp.pp_mesh_r / 0.5,
        ":",
        label="local + erf(r)/r",
    )
    # ax[0][0].plot(pp_mesh_r, (pp_mesh_r*pp_local+z_valence*erf(pp_mesh_r)/0.5)*np.sin(pp_mesh_r),
    #                ":", label="(r*local + erf(r)) sin(r)")
    ax[0][0].plot(
        pp.pp_mesh_r[15:],
        pp.pp_local[15:] + pp.z_valence / pp.pp_mesh_r[15:] / 0.5,
        ":",
        label="local + 1/r",
    )
    # ax[0][0].plot(pp_mesh_r[15:], pp_local[15:]-z_valence/pp_mesh_r[15:]/0.5,
    #                ":", label="local - 1/r")
    ax[0][0].plot(
        pp.pp_mesh_r,
        pp.pp_mesh_r * (pp.pp_mesh_r * pp.pp_local + pp.z_valence / 0.5),
        ":",
        label="r * (r * local + Z)",
    )
    ax[0][0].plot(
        pp.pp_mesh_r,
        pp.pp_mesh_r * (pp.pp_mesh_r * pp.pp_local + pp.z_valence / 0.5 * erf(pp.pp_mesh_r)),
        ":",
        label="r * (r * local + Z erf)",
    )
    # ax[0][0].plot(pp_mesh_r, pp_local*pp_mesh_r**2,
    #                ":", label="r^2 * local")
    # ax[0][0].plot(pp_mesh_r[40:], pp_local[40:]+(-z_valence/pp_mesh_r[40:]/0.5), label="local")
    ax[0][0].set_title("Local (long-ranged)")
    ax[0][0].legend()
    ax[0][0].grid()
    for key, val in pp.pp_nonlocal.items():
        if "BETA" in key:
            # r_i \beta(r_i) -> \beta(r_i)
            val = np.array(val) / np.array(pp.pp_mesh_r)
            ax[0][1].plot(pp.pp_mesh_r, pp.pp_mesh_r**2 * val, label=f"r^2 {key.split('_')[-1]}")  # , label=key.split(".")[-1])
            ax[0][1].set_title("Nonlocal Beta (short-ranged)")
            ax[0][1].legend()
            ax[0][1].grid(True)
        elif "DIJ" in key:
            # val = np.array(val).reshape((int(np.sqrt(len(val))),) * 2, order="C")
            im = ax[1][0].imshow(val, cmap="coolwarm")
            ax[1][0].set_title("Nonlocal D_ij")
            fig.colorbar(im)
        else:
            ax[1][0].plot(val, label=key)
            ax[1][0].set_title("Nonlocal D_ij")
            ax[1][0].grid()
    ax[1][1].plot(pp.pp_mesh_r, v_nl_r + pp.pp_local, label="v_nl + v_loc")
    ax[1][1].plot(pp.pp_mesh_r, -pp.z_valence / pp.pp_mesh_r / 0.5, "--", label="-Z/r")
    ax[1][1].plot(pp.pp_mesh_r, v_nl_r, ":", label="v_nl")
    ax[1][1].legend()
    ax[1][1].set_title("Full Nonlocal + Local")
    ax[1][1].set_yscale("symlog")
    ax[1][1].grid()

    # ax[1][1].plot(pp_rhoatom, label="rho atom")
    # ax[1][1].set_title("Rho Atom")
    # ax[1][1].grid()

    # for axis in ax.flatten():
    #     axis.grid()
    #     axis.legend()

    fig.show()

    # v_nl_mat = v_nl(
    #     wfc1_ncpp.k_plus_G,
    #     wfc1_ncpp.evc,
    #     wfc1_ncpp.cell_volume,
    #     wfc1_ncpp.atom_positions_hartree,
    #     pp,
    # )

    # v_loc_mat = v_loc(
    #     wfc1_ncpp.k_plus_G,
    #     wfc1_ncpp.evc,
    #     wfc1_ncpp.cell_volume,
    #     wfc1_ncpp.atom_positions_hartree,
    #     pp,
    # )

    atom_positions = wfc1_ncpp.atom_positions_hartree
    # atom_positions = np.array(
    #     [[0.0, -1., -4.0],[0.0, 1., 1.0], [0.0, -1., -1], [0.0, 1., 4]]
    # )
    atomic_numbers = wfc1_ncpp.atomic_numbers
    # atomic_numbers = np.array([4, 4, 1, 1])

    pps = calc_pps(
        p=wfc1_ncpp.k_plus_G,
        c_ip=wfc1_ncpp.evc,
        cell_volume=wfc1_ncpp.cell_volume,
        atom_positions=atom_positions,
        atomic_numbers=atomic_numbers,
        pseudopots=pseudopots,
    )

    # assert np.allclose(pps, v_nl_mat + v_loc_mat)

    import calc_matrix_elements

    iUj = calc_matrix_elements.iUj(
        wfc1_ncpp.k_plus_G,
        wfc1_ncpp.evc,
        atom_positions,
        atomic_numbers,
        wfc1_ncpp.cell_volume,
    )

    vmin = np.min([np.min(pps.real), np.min(-iUj.real)])
    vmax = np.max([np.max(pps.real), np.max(-iUj.real)])

    fig, ax = plt.subplots(ncols=3, figsize=(9, 3), constrained_layout=True)
    im0 = ax[0].imshow((pps).real, cmap="coolwarm", vmin=vmin, vmax=vmax)
    ax[0].set_title("PP")
    im1 = ax[1].imshow(-iUj.real, cmap="coolwarm", vmin=vmin, vmax=vmax)
    ax[1].set_title("iUj")
    im2 = ax[2].imshow(np.abs(-iUj - pps).real, cmap="coolwarm")
    ax[2].set_title("Diff.")
    fig.colorbar(im0)
    fig.colorbar(im1)
    fig.colorbar(im2)

    # plt.figure()
    # plt.title("iUj")
    # plt.imshow(iUj.real, cmap="coolwarm")
    # plt.colorbar()

    # plt.figure()
    # plt.title("Diff")
    # plt.imshow(np.abs(pps - iUj).real, cmap="coolwarm")
    # plt.colorbar()

    print(atom_positions)
    print(atomic_numbers)



import os
import sys
from dataclasses import dataclass
import xmltodict
import tomllib
import numpy as np
from dopyqo.colors import *
from dopyqo import units
from dopyqo.helpers.vqe_helpers import ExcitationPools, VQEOptimizers


@dataclass(frozen=True)
class DopyqoConfig:
    """Config for running a Dopyqo calculation

    Objects of this class are immutable (frozen instances) and variables of an object cannot be set manually after creation.
    Therefore, copying a object and changing values does not work. For this, use the dataclasses.replace function to create a new
    object from an existing one and simultaneously changing values.

    NOTE: None is the default of optional entries since we parse them as None from a toml-input file

    FIXME: If None is read from input file, just don't pass it into DopyqoConfig and set the defaults to non-None values?!

    Args:
        base_folder (str): Path to the folder in which the prefix.save folder is. For setting the prefix see argument prefix.
        prefix (str): Prefix of the Quantum ESPRESSO calculation. See also argument base_folder.
        active_electrons (int): Number of electrons in the active space.
        active_orbitals (int): Number of orbitals in the active space.
        kpoint_idx (int | str | None): Specify the k-point(s) that are calculated (integer or string).
                                       If integer than it specifies the index of the k-point in the k-point list used in Quantum ESPRESSO.
                                       If string than only "all" is allowed and all k-points are calculated. Set to 0, if None.
                                       Defaults to None.
        logging_flag (bool | None): If set to true, logging information will be shown. If None or False, no logging information will be shown. Defaults to None.
        n_threads (int | None): Number of threads used to calculate the pseudopotential when using the kohn-sham-matrix-elements-rs package. Set to 1 if None. Defaults to None.
        use_gpu (bool | None): If set to True, the GPU is used for ERI and frozen core matrix elements, if the cupy package is available.
                               If set to True and cupy package is not available, numpy will be used instead. Set to True if None. Defaults to None.
        wannier_transform (bool | None): If True, the orbitals in the active space are transformed into Wannier functions. See also wannier_umat and wannier_input_file.
                                         Set to False if None. Defaults to None.
        wannier_umat (str | None): Path to file holding the Wannier transformation matrix in Wannier90 format. Used only if wannier_transform is True.
                                   If None and wannier_transform is True will be set to base_folder/prefix_u.mat. Defaults to None.
        wannier_input_file (str | None): Path to the Wannier90 input file used to generate the u.mat file for argument wannier_umat.
                                         Used to validate if active space matches the transformed orbitals. Ignored if None. Defaults to None.
        occupations (list[int] | np.ndarray | None): NOT FULLY IMPLEMENTED, YET. List or numpy array of occupations used as the initial state in the VQE ansatz.
                                                     Only values of 0 and 1 are allowed. Defaults to None.
        run_vqe (bool | None): If set to True, a VQE calculation will be performed. Set to False if None. Defaults to None.
        run_fci (bool | None): If set to True, a FCI calculation will be performed. Set to False if None. Defaults to None.
        n_fci_energies (int | None): Number of energies calculated by the FCI solver. Set to 1 if None. Defaults to None.
        use_qiskit (bool | None): If set to True, qiskit is used to perform the VQE calculation. Set to False if None. Defaults to None.
        unit (units.Unit | None): Defines the unit used for the atom positions (see atom_positions) and lattice vectors (see lattice_vectors).
                                  Must not be None if either atom_positions or lattice_vectors are set. Defaults to None.
        atom_positions (np.ndarray | None): Set the atom positions with numpy array of shape (N, 3) where N is the number of atoms.
                                            Atoms are in the same order as in the output-xml file from Quantum ESPRESSO. Defaults to None.
        lattice_vectors (np.ndarray | None): Set lattice vectors with numpy array of shape (3, 3). Each line defines one lattice vector. Defaults to None.
        vqe_parameters (np.ndarray | None): Calculate energy of the VQE ansatz using fixed parameters. If given, no VQE optimization is performed.
                                            VQE optimization is performed if None. Defaults to None.
        vqe_initial_parameters (np.ndarray | None): Initial parameters of the VQE optimization. Set to zeros if None. Defaults to None.
        vqe_optimizer (VQEOptimizers | None): Optimizer used for the VQE optimization. Needs to be set if run_vqe is set to True. Defaults to None.
        vqe_maxiter (int | None): Number of VQE optimization steps before stopping the optimization. Set to large default values depending on the optimizer if None.
                                  Defaults to None.
        vqe_adapt (bool | None): If set to True, a APAPT-VQE calculation will be performed. If False, a VQE calculation will be performed. Set to False if None.
                                 Defaults to None.
        vqe_adapt_drain_pool (bool | None): If set to True the operator pool is drained when a operator is appended to the ansatz.
                                            If False, an operator can be appended to the ansatz multiple times. Set to True if None. Defaults to None.
        vqe_excitations (list[tuple[int, ...]] | ExcitationPools | None): Excitations used in the VQE ansatz our ADAPT-VQE pool.
                                                                          If given as list of tuples, each tuple represents one excitation.
                                                                          The tuples follow the notation used in TenCirChem:
                                                                          - An excitation operator is represented by a tuple of integers. Each integer corresponds to one spin-orbital.
                                                                          - The integers start from zero and denote first all up-spin orbitals then all down-spin orbitals
                                                                          - The first half of the tuple contains the indices for creation operator and the second half is for annilation operator.
                                                                          - Hermitian conjugation is handled internally.
                                                                           For example, (6, 2, 0, 4) corresponds to a†_6 a†_2 a_0 a_4 - a†_4 a†_0 a_2 a_6
                                                                          Set to dopyqo.ExcitationPools.SINGLES_DOUBLES if None. Defaults to None.
        uccsd_reps (int | None): Number of repetitions of the UCCSD ansatz used in VQE. Set to 1 if None. Defaults to None.
        qe_ewald (bool | None): If True, the nuclear-repulsion (Ewald) energy is not calculated but read from the xml-file outputted by Quantum ESPRESSO.
                                Set to False if None. Defaults to None.
        run_hf (bool | None): If True, a Hartree-Fock (HF) calculation is performed with PySCF and the Hamiltonian is transformed into the HF basis for the VQE/FCI.
                              Set to False if None. Defaults to None.
    """

    base_folder: str
    prefix: str
    active_electrons: int
    active_orbitals: int
    kpoint_idx: int | str | None = None
    logging_flag: bool | None = None
    n_threads: int | None = None
    use_gpu: bool | None = None
    wannier_transform: bool | None = None
    wannier_umat: str | None = None
    wannier_input_file: str | None = None
    occupations: list[int] | np.ndarray | None = None
    run_vqe: bool | None = None
    run_fci: bool | None = None
    n_fci_energies: int | None = None  # TODO: Add this to toml-input-file
    use_qiskit: bool | None = None
    unit: units.Unit | None = None
    atom_positions: np.ndarray | None = None
    lattice_vectors: np.ndarray | None = None
    vqe_parameters: np.ndarray | None = None  # No VQE optimization, just run circuit with these parameters
    vqe_initial_parameters: np.ndarray | None = None  # TODO: Add this to toml-input-file
    vqe_optimizer: VQEOptimizers | None = None
    vqe_maxiter: int | None = None
    vqe_adapt: bool | None = None
    vqe_adapt_drain_pool: bool | None = None
    vqe_excitations: list[tuple[int, ...]] | ExcitationPools | None = None
    uccsd_reps: int | None = None
    qe_ewald: bool | None = None
    run_hf: bool | None = None  # TODO: Add this to toml-input-file

    def __post_init__(self):
        if len(self.base_folder) > 0 and not os.path.isdir(self.base_folder):
            print(f"{RED}Config error: Base folder ({self.base_folder}) does not exist!{RESET_COLOR}", file=sys.stderr)
            sys.exit(1)
        if not os.path.isdir(os.path.join(self.base_folder, f"{self.prefix}.save")):
            print(
                f"{RED}Config error: QE save folder ({os.path.join(self.base_folder, f'{self.prefix}.save')}) does not exist!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if self.active_electrons is None:
            print(f"{RED}Config error: Number of active electrons (active_electrons) has to be set but is not!{RESET_COLOR}", file=sys.stderr)
            sys.exit(1)
        if self.active_orbitals is None:
            print(f"{RED}Config error: Number of active orbitals (active_orbitals) has to be set but is not!{RESET_COLOR}", file=sys.stderr)
            sys.exit(1)
        if self.kpoint_idx is None:
            object.__setattr__(self, "kpoint_idx", 0)
        if (not isinstance(self.kpoint_idx, int) or isinstance(self.kpoint_idx, str)) and self.kpoint_idx != "all":
            print(
                f'{RED}Config error: Invalid k-point index ({self.kpoint_idx})! Either set to integer starting from zero or set to "all"!{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        if isinstance(self.kpoint_idx, int) and self.kpoint_idx < 0:
            print(
                f'{RED}Config error: Given k-point index ("{self.kpoint_idx}") negative but needs to be positive!{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        n_wfc_files = len([x for x in os.listdir(os.path.join(self.base_folder, f"{self.prefix}.save")) if x.startswith("wfc")])
        if isinstance(self.kpoint_idx, int) and self.kpoint_idx + 1 > n_wfc_files:
            print(
                f"{RED}Config error: Given k-point index ({self.kpoint_idx}, starting at zero) is to large! Found only {n_wfc_files} wavefunction (wfc) files!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        xml_file = os.path.join(self.base_folder, f"{self.prefix}.save", "data-file-schema.xml")
        with open(xml_file, encoding="utf-8") as file:
            xml_dict = xmltodict.parse(file.read())
        n_kpoints = int(xml_dict["qes:espresso"]["output"]["band_structure"]["nks"])
        if n_kpoints != 1 and self.kpoint_idx is None:
            print(
                f"{RED}Config error: You try loading a DFT calculation involving multiple k-points ({self.n_kpoints}) "
                + "without specifying the k-point you want to load (kpoint_idx=None). "
                + "Specify a k-point by setting kpoint_idx to the corresponding index of the k-point, starting at zero. "
                + f'You can also load all k-points by setting kpoint_idx to "all"!{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        if self.run_vqe is None:
            object.__setattr__(self, "run_vqe", False)
        if self.run_fci is None:
            object.__setattr__(self, "run_fci", False)
        if self.run_fci and self.n_fci_energies is None:
            object.__setattr__(self, "n_fci_energies", 1)
        if self.atom_positions is not None:
            if self.unit is None:
                allowed_units_str = "\n\t".join(units.ALLOWED_UNITS)
                print(
                    f"{RED}Config error: Atom positions defined but unit not defined. Please define unit! Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        if self.lattice_vectors is not None:
            if self.lattice_vectors.shape != (3, 3):
                print(
                    f"{RED}Config error: Lattice vectors are invalid. Expected a numpy array with shape (3, 3) but got {self.lattice_vectors.shape}.{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if self.unit is None:
                allowed_units_str = "\n\t".join(units.ALLOWED_UNITS_LATTICE)
                print(
                    f"{RED}Config error: Atom positions defined but unit not defined. Please define unit! Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if self.unit.name.lower() not in units.ALLOWED_UNITS_LATTICE:
                allowed_units_str = "\n\t".join(units.ALLOWED_UNITS_LATTICE)
                print(
                    f"{RED}Config error: Defined unit ({self.unit}) cannot be used when defining lattice vectors. Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        if self.wannier_umat is not None and not self.wannier_transform:
            print(
                f"{ORANGE}Config warning: Wannier U-matrix path (wannier_umat) set but no Wannier transformation will be performed (wannier_transform not True).{RESET_COLOR}",
                file=sys.stdout,
            )
        if self.wannier_input_file is not None and not self.wannier_transform:
            print(
                f"{ORANGE}Config warning: Wannier input file path (wannier_input_file) set but no Wannier transformation will be performed (wannier_transform not True).{RESET_COLOR}",
                file=sys.stdout,
            )
        if self.wannier_input_file is not None and not os.path.isfile(self.wannier_input_file):
            print(
                f"{RED}Config error: Wannier input file ({self.wannier_input_file}) does not exist!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)

        if self.vqe_adapt is not None and self.vqe_adapt:
            if self.uccsd_reps is not None:
                print(
                    f"{ORANGE}Config warning: Both "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" and "
                    + f"{self.uccsd_reps=}".split("=")[0][5:]
                    + f" are set. "
                    + f"{self.uccsd_reps=}".split("=")[0][5:]
                    + f" has no effect and will be ignored!{RESET_COLOR}",
                )
            if self.vqe_initial_parameters is not None:
                print(
                    f"{ORANGE}Config warning: Both "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" and "
                    + f"{self.vqe_initial_parameters=}".split("=")[0][5:]
                    + f" are set. "
                    + f"{self.vqe_initial_parameters=}".split("=")[0][5:]
                    + f" has no effect and will be ignored!{RESET_COLOR}",
                )
        if self.vqe_optimizer is None and self.run_vqe:
            print(
                f"{RED}Config error: "
                + f"{self.run_vqe=}".split("=")[0][5:]
                + f" is True but "
                + f"{self.vqe_optimizer=}".split("=")[0][5:]
                + f"is None but has to be set if "
                + f"{self.run_vqe=}".split("=")[0][5:]
                + f" is True!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if not isinstance(self.vqe_optimizer, VQEOptimizers) and self.vqe_optimizer is not None:
            print(
                f"{RED}Config error: "
                + f"{self.vqe_optimizer=}".split("=")[0][5:]
                + f" has to be of type dopyqo.VQEOptimizers but is of type {type(self.vqe_optimizer)}!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if self.uccsd_reps is None:
            object.__setattr__(self, "uccsd_reps", 1)
        if self.qe_ewald is None:
            object.__setattr__(self, "qe_ewald", False)
        if self.run_hf is None:
            object.__setattr__(self, "pyscf_hf", False)
        if self.n_threads is None or self.n_threads <= 0:
            object.__setattr__(self, "n_threads", 1)
        if self.use_gpu is None:
            object.__setattr__(self, "use_gpu", True)
        if self.vqe_adapt is None:
            object.__setattr__(self, "vqe_adapt", False)
        if self.vqe_excitations is None:
            object.__setattr__(self, "vqe_excitations", ExcitationPools.SINGLES_DOUBLES)
        if not isinstance(self.vqe_excitations, list) and not isinstance(self.vqe_excitations, ExcitationPools):
            print(
                f"{RED}Config error: VQE excitations parameter has to be of type list or of type dopyqo.ExcitationPools but is of type {type(self.vqe_excitations)}!{RESET_COLOR}",
                file=sys.stderr,
            )
            sys.exit(1)
        if self.vqe_adapt:
            if not self.run_vqe:
                print(
                    f"{ORANGE}Config warning: "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" is set to True but "
                    + f"{self.run_vqe=}".split("=")[0][5:]
                    + f" is set to False. "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" has no effect and will be ignored!{RESET_COLOR}",
                )
        if self.vqe_adapt_drain_pool is not None and self.vqe_adapt_drain_pool:
            if not self.vqe_adapt:
                print(
                    f"{ORANGE}Config warning: "
                    + f"{self.vqe_adapt_drain_pool=}".split("=")[0][5:]
                    + f" is set to True but "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" is set to False. "
                    + f"{self.vqe_adapt_drain_pool=}".split("=")[0][5:]
                    + f" has no effect and will be ignored!{RESET_COLOR}",
                )
        if self.vqe_adapt_drain_pool is None:
            if self.vqe_adapt:
                object.__setattr__(self, "vqe_adapt_drain_pool", True)
            else:
                object.__setattr__(self, "vqe_adapt_drain_pool", False)
        if self.vqe_excitations is not None:
            if self.use_qiskit:
                print(
                    f"{ORANGE}Config warning: VQE excitations set but Qiskit will be used. Costum VQE excitations currently only supported when using TenCirChem. Given VQE excitations will be ignored!{RESET_COLOR}"
                )
            if isinstance(self.vqe_excitations, list):
                for exc in self.vqe_excitations:
                    if not isinstance(exc, tuple):
                        print(
                            f"{RED}Config error: VQE excitations parameter has to be a list of tuples but is a list that contains elements of type {type(exc)}!{RESET_COLOR}",
                            file=sys.stderr,
                        )
                        sys.exit(1)
                    for idx in exc:
                        if not isinstance(idx, int):
                            print(
                                f"{RED}Config error: VQE excitations parameter has to be a list of tuples of ints but is a list of tuples coantining elements of type {type(idx)} in tuple {exc}!{RESET_COLOR}",
                                file=sys.stderr,
                            )
                            sys.exit(1)
                    if len(exc) % 2 != 0:
                        print(
                            f"{RED}Config error: VQE excitations parameter has to be a list of tuples of an even number of ints but found tuple of length {len(exc)} (tuple: {exc})!{RESET_COLOR}",
                            file=sys.stderr,
                        )
                        sys.exit(1)
        if self.vqe_adapt:
            if self.use_qiskit:
                print(
                    f"{RED}Config error: Both "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" and "
                    + f"{self.use_qiskit=}".split("=")[0][5:]
                    + f" are set to True. ADAPT-VQE is not supported when using Qiskit!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        if self.vqe_parameters is not None:
            if self.vqe_adapt:
                print(
                    f"{RED}Config error: VQE parameters given by "
                    + f"{self.vqe_parameters=}".split("=")[0][5:]
                    + f" but "
                    + f"{self.vqe_adapt=}".split("=")[0][5:]
                    + f" is set to True. "
                    + f"{self.vqe_parameters=}".split("=")[0][5:]
                    + f" can only be set for non-ADAPT VQE calculations!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if isinstance(self.vqe_excitations, list):
                n_exc = len(self.vqe_excitations)
            elif isinstance(self.vqe_excitations, ExcitationPools):
                print(
                    f"{RED}Config error: Type dopyqo.ExcitationPools currently not supported for "
                    + f"{self.vqe_excitations=}".split("=")[0][5:]
                    + f" if  "
                    + f"{self.vqe_parameters=}".split("=")[0][5:]
                    + f" are given. Provide a list of excitations instead!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            else:
                no = self.active_electrons // 2
                norb = self.active_orbitals
                nv = norb - no

                n_singles = no * nv * 2
                n_doubles = 2 * (no * (no - 1) // 2) * (nv * (nv - 1) // 2)
                for i in range(no):
                    for j in range(i + 1):
                        for a in range(nv):
                            for b in range(a + 1):
                                if i == j and a == b:
                                    n_doubles += 1
                                    continue
                                n_doubles += 2
                                if (i != j) and (a != b):
                                    n_doubles += 2
                n_exc = (n_singles + n_doubles) * self.uccsd_reps
            if len(self.vqe_parameters) != n_exc:
                print(
                    f"{RED}Config error: Expected {n_exc} VQE parameters but got {len(self.vqe_parameters)}!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        if self.vqe_initial_parameters is not None:
            if self.vqe_parameters is not None:
                print(f"{RED}Config error: vqe_initial_parameters set but also vqe_parameters is set. Choose one!{RESET_COLOR}", file=sys.stderr)
                sys.exit(1)

            if self.vqe_excitations is not None:
                n_exc = len(self.vqe_excitations)
            else:
                no = self.active_electrons // 2
                norb = self.active_orbitals
                nv = norb - no

                n_singles = no * nv * 2
                n_doubles = 2 * (no * (no - 1) // 2) * (nv * (nv - 1) // 2)
                for i in range(no):
                    for j in range(i + 1):
                        for a in range(nv):
                            for b in range(a + 1):
                                if i == j and a == b:
                                    n_doubles += 1
                                    continue
                                n_doubles += 2
                                if (i != j) and (a != b):
                                    n_doubles += 2
                n_exc = (n_singles + n_doubles) * self.uccsd_reps
            if len(self.vqe_initial_parameters) != n_exc:
                print(
                    f"{RED}Config error: Expected {n_exc} VQE initial parameters but got {len(self.vqe_initial_parameters)}!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        if self.wannier_transform:
            if self.wannier_umat is None:
                object.__setattr__(self, "wannier_umat", os.path.join(self.base_folder, f"{self.prefix}_u.mat"))
            if not os.path.isfile(self.wannier_umat):
                print(
                    f"{RED}Config error: Wannier U-matrix ({self.wannier_umat}) does not exist!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if self.wannier_input_file is None:
                object.__setattr__(self, "wannier_input_file", os.path.join(self.base_folder, f"{self.prefix}.win"))
            if not os.path.isfile(self.wannier_input_file):
                print(
                    f"{ORANGE}Config warning: Wannier input file (automatically set to {self.wannier_input_file}) does not exist! ",
                    f"If the Wannier-transformed orbitals match the orbitals in the active space cannot be checked.{RESET_COLOR}",
                )
        if self.occupations is not None:
            if not self.run_vqe:
                print(
                    f"{ORANGE}Config warning: occupations are specified but no VQE calculation will be performed. occupations will not be used!{RESET_COLOR}"
                )
            if self.use_qiskit:
                print(
                    f"{ORANGE}Config warning: occupations are specified but qiskit will be used for the VQE calculation. occupations are currently only supported when using TenCirChem. occupations will not be used!{RESET_COLOR}"
                )
            if not isinstance(self.occupations, list) and not isinstance(self.occupations, np.ndarray):
                print(
                    f"{RED}Config error: occupations has to be a list or numpy array, but is {type(self.occupations)}!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if isinstance(self.occupations, np.ndarray) and self.occupations.size != self.active_orbitals:
                print(
                    f"{RED}Config error: occupations must have {self.active_orbitals} elements (number of orbitals) but has {self.occupations.size} elements (shape {self.occupations.shape})!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            if len(self.occupations) != self.active_orbitals:
                print(
                    f"{RED}Config error: occupations must have {self.active_orbitals} elements (number of orbitals) but has {len(self.occupations)} elements!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
            for x in self.occupations:
                if not np.isclose(x, int(x)):
                    print(
                        f"{RED}Config error: occupations must be integers but found value {x}!{RESET_COLOR}",
                        file=sys.stderr,
                    )
                    sys.exit(1)
            object.__setattr__(self, "occupations", np.array([int(x) for x in self.occupations]))
            if not np.isclose(2.0 * np.sum(self.occupations), self.active_electrons):
                print(
                    f"{RED}Config error: Two times the occupations must sum to the number of electrons ({self.active_electrons}) but sum to {2.0*np.sum(self.occupations)}!{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)


def read_control(control: dict) -> tuple:
    base_folder = control.get("base_folder")
    prefix = control.get("prefix")
    active_electrons = control.get("active_electrons")
    active_orbitals = control.get("active_orbitals")
    kpoint_idx = control.get("kpoint_idx")
    logging_flag = control.get("logging")
    n_threads = control.get("n_threads")
    run_vqe = control.get("run_vqe")
    run_fci = control.get("run_fci")
    use_qiskit = control.get("use_qiskit")

    if active_electrons is None:
        print(
            f'{RED}Error in input file: Number of active electrons ("active_electrons") has to be defined in section "control" but is not!{RESET_COLOR}',
            file=sys.stderr,
        )
        sys.exit(1)
    if active_orbitals is None:
        print(
            f'{RED}Error in input file: Number of active orbitals ("active_orbitals") has to be defined in section "control" but is not!{RESET_COLOR}',
            file=sys.stderr,
        )
        sys.exit(1)

    return base_folder, prefix, active_electrons, active_orbitals, kpoint_idx, logging_flag, n_threads, run_vqe, run_fci, use_qiskit


def read_wannier(wannier: dict) -> tuple:
    wannier_transform = wannier.get("transform")
    wannier_umat = wannier.get("umat")
    wannier_input_file = wannier.get("input_file")

    return wannier_transform, wannier_umat, wannier_input_file


def read_geometry(geometry: dict) -> tuple[units.Unit | None, np.ndarray | None, np.ndarray | None]:
    allowed_units_str = "\n\t".join(units.ALLOWED_UNITS)
    unit = None

    def get_unit(unit: str) -> units.Unit:
        match unit:
            case "angstrom":
                unit = units.Unit.ANGSTROM
            case "bohr":
                unit = units.Unit.HARTREE
            case "meter":
                unit = units.Unit.METER
            case "alat":
                unit = units.Unit.ALAT
            case "crystal":
                unit = units.Unit.CRYSTAL
            case _:
                print(
                    f"{RED}Error in input file: Unit {unit} not supported! Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}",
                    file=sys.stderr,
                )
                sys.exit(1)
        return unit

    ######################### COORDINATES #########################
    # IDEA: dict instead of list? Element: coordinates e.g. {"H":[0.0, 0.1, 1.5]}?
    atom_positions = geometry.get("coordinates")
    if atom_positions is not None:
        if "unit" not in geometry.keys():
            print(
                f'{RED}Error in input file: Atom positions defined in section "coordinates" but "unit" not defined. Please define "unit"! Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        unit = get_unit(str(geometry["unit"]).lower())
        atom_positions = np.array(atom_positions)

    ######################### LATTICE_VECTORS #########################
    allowed_units_lattice_str = "\n\t".join(units.ALLOWED_UNITS_LATTICE)
    lattice_vectors = geometry.get("lattice_vectors")
    if lattice_vectors is not None:
        lattice_vectors = np.array(lattice_vectors)
        if lattice_vectors.shape != (3, 3):
            print(
                f'{RED}Error in input file: Lattice vectors defined in section "lattice_vectors" are invalid. Expected a list of lists with shape (3, 3) but got {lattice_vectors.shape}.{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        if "unit" not in geometry.keys():
            print(
                f'{RED}Error in input file: Lattice vectors defined in section "lattice_vectors" but "unit" not defined. Please define "unit"! Supported units are:\n\t{allowed_units_str}\n{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)
        unit = get_unit(str(geometry["unit"]).lower())
        if unit.name.lower() not in units.ALLOWED_UNITS_LATTICE:
            print(
                f'{RED}Error in input file: Defined unit ({geometry["unit"]}) cannot be used when defining lattice vectors in section "lattice_vectors". Supported units are:\n\t{allowed_units_lattice_str}\n{RESET_COLOR}',
                file=sys.stderr,
            )
            sys.exit(1)

    return unit, atom_positions, lattice_vectors


def read_vqe(vqe: dict) -> tuple[np.ndarray | None, str | None, int | None]:
    ######################### PARAMETERS #########################
    parameters = vqe.get("parameters")
    parameters = np.array(parameters) if parameters is not None else None

    ######################### OPTIMIZER #########################
    optimizer = vqe.get("optimizer")
    if optimizer is not None:
        optimizer = str(optimizer)
        if optimizer.lower() == "l-bfgs-b":
            optimizer = VQEOptimizers.L_BFGS_B
        elif optimizer.lower() == "cobyla":
            optimizer = VQEOptimizers.COBYLA
        elif optimizer.lower() == "excitationsolve":
            optimizer = VQEOptimizers.ExcitationSolve
        else:
            allowed_opt = [x.name.lower() for x in VQEOptimizers]
            print(f"{RED}VQE optimizer {optimizer} is not supported. Please use on of {allowed_opt}!{RESET_COLOR}")

    ######################### UCCSD_REPS #########################
    uccsd_reps = vqe.get("uccsd_reps")
    uccsd_reps = int(uccsd_reps) if uccsd_reps is not None else None

    return parameters, optimizer, uccsd_reps


def read_input_toml(file) -> DopyqoConfig:
    allowed_sections = ["control", "wannier", "geometry", "vqe"]
    with open(file, "rb") as f:
        config_toml = tomllib.load(f)

    present_sections = config_toml.keys()
    unallowed_sections_present = False
    for sec_tmp in present_sections:
        if sec_tmp not in allowed_sections:
            print(f"Section {sec_tmp} in TOML-file is not supported and ignored!")
            unallowed_sections_present = True
    if unallowed_sections_present:
        allowed_str_tmp = "\n\t".join(allowed_sections)
        print(f"Supported sections are:\n\t{allowed_str_tmp}\n")

    config_dict = {}
    ######################### CONTROL #########################
    control = config_toml.get("control")
    if control is not None:
        base_folder, prefix, active_electrons, active_orbitals, kpoint_idx, logging_flag, n_threads, run_vqe, run_fci, use_qiskit = read_control(
            control
        )
        config_dict["base_folder"] = base_folder
        config_dict["prefix"] = prefix
        config_dict["active_electrons"] = active_electrons
        config_dict["active_orbitals"] = active_orbitals
        config_dict["logging_flag"] = logging_flag
        config_dict["n_threads"] = n_threads
        config_dict["kpoint_idx"] = kpoint_idx
        config_dict["run_vqe"] = run_vqe
        config_dict["run_fci"] = run_fci
        config_dict["use_qiskit"] = use_qiskit

    ######################### WANNIER #########################
    wannier = config_toml.get("wannier")
    if wannier is not None:
        wannier_transform, wannier_umat, wannier_input_file = read_wannier(wannier)
        config_dict["wannier_transform"] = wannier_transform
        config_dict["wannier_umat"] = wannier_umat
        config_dict["wannier_input_file"] = wannier_input_file

    ######################### GEOMETRY #########################
    geometry = config_toml.get("geometry")
    if geometry is not None:
        unit, atom_positions, lattice_vectors = read_geometry(geometry)
        config_dict["unit"] = unit
        config_dict["atom_positions"] = atom_positions
        config_dict["lattice_vectors"] = lattice_vectors

    ######################### VQE #########################
    vqe = config_toml.get("vqe")
    if vqe is not None:
        parameters, optimizer, uccsd_reps = read_vqe(vqe)
        config_dict["vqe_parameters"] = parameters
        config_dict["vqe_optimizer"] = optimizer
        config_dict["uccsd_reps"] = uccsd_reps

    return config_dict



import os
import numpy as np
import matplotlib.pyplot as plt
from tencirchem.static.hamiltonian import random_integral
import dopyqo
from dopyqo.colors import *

if __name__ == "__main__":
    # Defining custom core potential and custom electron-electron interaction
    active_orbitals = 4
    h1e_core_custom, h2e_custom = random_integral(active_orbitals)
    h2e_custom = h2e_custom.transpose((0, 2, 3, 1))  # Tranform from chemist index order to physicist index order

    mats_custom = dopyqo.MatrixElements(
        h_pq_core=h1e_core_custom,
        h_pqrs=h2e_custom,
        # Other matrix elements and energies are automatically set to None and are calculated
    )

    vqe_optimizer = dopyqo.VQEOptimizers.L_BFGS_B
    config = dopyqo.DopyqoConfig(
        base_folder=os.path.join("qe_files", "Be"),
        prefix="Be",
        active_electrons=2,
        active_orbitals=active_orbitals,
        run_fci=True,  # Run FCI calculation
        run_vqe=True,  # Run VQE calculation
        vqe_optimizer=vqe_optimizer,
        vqe_excitations=dopyqo.ExcitationPools.SINGLES_DOUBLES,
        n_threads=10,
    )
    energy_dict, wfc_obj, h_ks, mats = dopyqo.run(config, mats_custom)
    dft_energy = energy_dict["dft_energy"]
    fci_energy = energy_dict["fci_energy"]
    vqe_energy = energy_dict["vqe_energy"]

    plt.plot(h_ks.vqe_counts, np.abs(h_ks.vqe_values - fci_energy), linestyle="-", marker="x")
    plt.yscale("log")
    plt.grid()



import logging
import os
import xmltodict
from dataclasses import dataclass
import numpy as np
import dopyqo


@dataclass
class BetaProjector:
    r"""Projector r \beta^i(r) (note the factor r)

    Args:
        idx (int): index i of projector starting at 1. Written as PP_BETA.idx in PP file
        angular_momentum (int): Angular momentum value of projector
        values (np.ndarray): Values of r \beta^i(r) for different values of r (note the factor r)
    """

    idx: int
    angular_momentum: int
    values: np.ndarray


@dataclass
class rGrid:
    r"""A real-space grid where a pseudopotential is defined on

    Args:
        r (np.ndarray): radial grid points
        rab (np.ndarray): factor required for discrete integration: \int f(r) dr = \sum_i f_i rab_i.
    """

    def __init__(self, r: np.ndarray, rab: np.ndarray):
        assert (lr := len(r)) == (
            lrab := len(rab)
        ), f"Real-space grid invalid. Number radial grid point ({lr}) is not equal to number of rab factors ({lrab})!"

        self.r = r
        self.rab = rab


class Pseudopot:
    r"""Pseudopotential (PP) class holding information from the PP file in RYDBERG atomic units

    Args:
        pp_file (str): Path to the pseudopotential file
    """

    def __init__(self, pp_file: str):
        with open(pp_file, "r", encoding="utf-8") as file:
            xml_dict = xmltodict.parse(file.read())["UPF"]

        self.version = xml_dict["@version"]
        self.pp_info = xml_dict["PP_INFO"]
        self.pp_input = self.pp_info["PP_INPUTFILE"]
        # ref_inp_columns = [
        #     "#",
        #     "atsym",
        #     "z",
        #     "nc",
        #     "nv",
        #     "iexc",
        #     "psfile",
        # ]
        # assert (
        #     pp_input.split("\n")[1].split() == ref_inp_columns
        # ), "Cannot read information about atom in pseudopotential!"
        # assert len(pp_input.split("\n")[2].split()) == len(ref_inp_columns) - 1
        # atsym, z, nc, nv, iexc, psfile = pp_input.split("\n")[2].split()

        self.pp_header = xml_dict["PP_HEADER"]
        self.element = self.pp_header["@element"].strip()
        self.atomic_number = dopyqo.elements_to_atomic_number[self.element]
        self.z_valence = float(self.pp_header["@z_valence"])
        self.z_core = self.atomic_number - self.z_valence
        self.n_proj = int(self.pp_header["@number_of_proj"])

        logging.info(f"Pseudopotential for {self.element} with core of {int(self.z_core)} electrons.")

        assert self.pp_header["@pseudo_type"] == "NC", f"Only fully nonlocal NCPPs supported but found {self.pp_header['@pseudo_type']} PP!"
        assert self.pp_header["@is_coulomb"] == "F", "Only non-coulombic potentials supported but PP is coulombic!"

        self.pp_mesh = xml_dict["PP_MESH"]
        self.pp_mesh_r = np.array(list(map(float, self.pp_mesh["PP_R"]["#text"].split())))
        # \int f(r) dr = \sum_i f_i rab_i
        self.pp_mesh_rab = np.array(list(map(float, self.pp_mesh["PP_RAB"]["#text"].split())))
        self.r_grid = rGrid(self.pp_mesh_r, self.pp_mesh_rab)

        # NOTE: The loaded potentials are centered at the origin. To have them atom-centered
        #       use \vec{r} -> \vec{r} - \vec{R}, where \vec{R} is the position of the atom.
        #       The local potential is only dependent on the norm so |r| -> |r-R|

        # local potential (Ry a.u.) sampled on the radial grid
        # is spherical symmetric and does not involve e.g. spherical harmonics,
        # i.e. V_loc(\vec{r}) = V_loc(|r|)
        self.pp_local = np.array(list(map(float, xml_dict["PP_LOCAL"]["#text"].split())))

        # BETA: projector r_i \beta(r_i) (note the factor r)
        # DIJ:  Dij factors of the nonlocal PP in row-major order:
        #       V_{NL} = \sum_{i,j} D_{i,j} |\beta_i><\beta_j|
        # Angular part is described by spherical harmonics
        self.pp_nonlocal = {}
        for key, val in xml_dict["PP_NONLOCAL"].items():
            self.pp_nonlocal[key] = np.array(list(map(float, val["#text"].split())))
        # Save Dij in numpy array using row-major order
        self.pp_nonlocal["PP_DIJ"] = np.array(self.pp_nonlocal["PP_DIJ"]).reshape((int(np.sqrt(len(self.pp_nonlocal["PP_DIJ"]))),) * 2, order="C")
        self.pp_betas = [  # Same as pp_nonlocal["PP_BETA.x"] but with angular momentum information
            BetaProjector(
                idx=int(key.split(".")[-1]),
                angular_momentum=int(val["@angular_momentum"]),
                values=np.array(list(map(float, val["#text"].split()))),
            )
            for key, val in xml_dict["PP_NONLOCAL"].items()
            if "BETA" in key  # and int(key.split(".")[-1]) <= n_proj
        ]

        self.pp_dij = self.pp_nonlocal["PP_DIJ"]

        # Non-local potential V_NL
        # V_NL(r) = \sum_{l,m} \sum_{i,j} \beta_{i,l}(r) Y_{lm}(r) D_{i,j} \beta*_{j,l}(r) Y*_{lm}(r)
        # self.v_nl_r = np.zeros_like(self.pp_mesh_r)
        # for i in range(self.pp_nonlocal["PP_DIJ"].shape[0]):
        #     for j in range(self.pp_nonlocal["PP_DIJ"].shape[1]):
        #         self.v_nl_r += (
        #             self.pp_nonlocal["PP_DIJ"][i, j]
        #             * self.pp_nonlocal[f"PP_BETA.{i+1}"]
        #             / np.array(self.pp_mesh_r)
        #             * self.pp_nonlocal[f"PP_BETA.{j+1}"]
        #             / np.array(self.pp_mesh_r)
        #         )

        self.pp_pswfc = xml_dict["PP_PSWFC"]

        # radial atomic (pseudo-)charge. This is 4\pi r^2 times the true charge.
        self.pp_rhoatom = list(map(float, xml_dict["PP_RHOATOM"]["#text"].split()))


if __name__ == "__main__":

    xml_file = os.path.join("qe_files", "pseudo", "H_ONCV_PBE-1.2.upf")
    pp = Pseudopot(xml_file)
    print(pp.__dict__)



import os
import numpy as np
import dopyqo
from dopyqo.colors import *

if __name__ == "__main__":
    config = dopyqo.DopyqoConfig(
        # base_folder=os.path.join("qe_files", "Be"),
        # prefix="Be",
        base_folder=os.path.join("qe_files", "Mg2Si"),
        prefix="Mg2Si",
        active_electrons=2,
        active_orbitals=2,
        n_threads=10,
    )

    # # We could run a dopyqo calculation, which would also calculate the matrix elements, but we just need the Wfc object for plotting
    # energy_dict, wfc_obj, h, mats = run_dopyqo(config, show_banner=False, verbosity=0)

    # Calculating the matrix elements is not necessary, we just need the Wfc object. Let's create this manually!
    xml_file = os.path.join(config.base_folder, f"{config.prefix}.save", "data-file-schema.xml")
    wfc_file = os.path.join(config.base_folder, f"{config.prefix}.save", "wfc1.dat")
    wfc_obj = dopyqo.Wfc.from_file(wfc_file, xml_file, kpoint_idx=config.kpoint_idx)

    plotter = wfc_obj.plot_real_space(
        band_idc=[8],
        isosurfaces=5,
        opacity=0.6,
        extend_data=False,
        extend_atoms=False,
        plot_lattice_vectors=True,
        html_filename=os.path.join(config.base_folder, "plot_wfc_ks"),
    )
